"""Example showing how to produce an Arrow table from DuckDB."""

import duckdb
import pyarrow as pa
import ctypes

# Execute a query and fetch results as Arrow
con = duckdb.connect()
record_batch = con.execute("SELECT 1 AS a, 'foo' AS b").fetch_record_batch()

# Export as an Arrow C Stream pointer for zero-copy transfer
reader = pa.RecordBatchReader.from_batches(record_batch.schema, [record_batch])
if hasattr(reader, "__arrow_c_stream__"):
    capsule = reader.__arrow_c_stream__()
    stream_ptr = ctypes.cast(capsule, ctypes.c_void_p).value
else:
    from pyarrow.cffi import ffi
    c_stream = ffi.new("struct ArrowArrayStream*")
    reader._export_to_c(c_stream)
    stream_ptr = int(ffi.cast("uintptr_t", c_stream))

print("Arrow stream pointer:", stream_ptr)



import multiprocessing
import socket
import subprocess
import sys
import time
from pathlib import Path

import psycopg
import unittest
from helpers import _ensure_riffq_built

def _run_server(port: int):
    import riffq
    from riffq.helpers import to_arrow

    def handle_query(sql, callback, **kwargs):
        callback(to_arrow([{"name": "val", "type": "int"}], [[1]]))

    def handle_auth(conn_id, user, password, host, *, callback, database=None):
        callback(user == "user" and password == "secret")

    server = riffq.Server(f"127.0.0.1:{port}")
    server.on_query(handle_query)
    server.on_authentication(handle_auth)
    server.start()


class AuthenticationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _ensure_riffq_built()
        cls.port = 55438
        cls.proc = multiprocessing.Process(target=_run_server, args=(cls.port,), daemon=True)
        cls.proc.start()
        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()

    def test_auth_reject(self):
        with self.assertRaises(psycopg.OperationalError):
            psycopg.connect(f"postgresql://user:wrong@127.0.0.1:{self.port}/db")

    def test_auth_accept(self):
        conn = psycopg.connect(f"postgresql://user:secret@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT 1")
            self.assertEqual(cur.fetchone()[0], 1)
        conn.close()


if __name__ == "__main__":
    unittest.main()



import logging
import duckdb
import pyarrow as pa
import riffq
logging.basicConfig(level=logging.DEBUG)

class Connection(riffq.BaseConnection):
    def _handle_query(self, sql, callback, **kwargs):
        cur = duckdb_con.cursor()
        text = sql.strip().lower().split(';')[0]

        if text == "select pg_catalog.version()":
            batch = self.arrow_batch(
                [pa.array(["PostgreSQL 14.13"])],
                ["version"],
            )
            return self.send_reader(batch, callback)

        if text == "show transaction isolation level":
            batch = self.arrow_batch(
                [pa.array(["read committed"])],
                ["transaction_isolation"],
            )
            return self.send_reader(batch, callback)

        if text == "show standard_conforming_strings":
            batch = self.arrow_batch(
                [pa.array(["read committed"])],
                ["transaction_isolation"],
            )
            return self.send_reader(batch, callback)
        
        if text == "select current_schema()":
            batch = self.arrow_batch(
                [pa.array(["public"])],
                ["current_schema"],
            )
            return self.send_reader(batch, callback)

        try:
            reader = cur.execute(sql).fetch_record_batch()
            self.send_reader(reader, callback)
        except Exception as exc:
            logging.exception("error on executing query")
            batch = self.arrow_batch(
                [pa.array(["ERROR"]), pa.array([str(exc)])],
                ["error", "message"],
            )
            self.send_reader(batch, callback)

    def handle_query(self, sql, callback=callable, **kwargs):
        self.executor.submit(self._handle_query, sql, callback, **kwargs)

def main():
    global duckdb_con
    duckdb_con = duckdb.connect()
    duckdb_con.execute(
        """
        CREATE VIEW klines AS 
        SELECT * 
        FROM 'data/klines.parquet'
        """
    )
    server = riffq.RiffqServer("127.0.0.1:5433", connection_cls=Connection)
    server.set_tls("certs/server.crt", "certs/server.key")
    server.start()

if __name__ == "__main__":
    main()



import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime

import duckdb
import pyarrow as pa
import riffq

def send_reader(reader, callback):
    if hasattr(reader, "__arrow_c_stream__"):
        capsule = reader.__arrow_c_stream__()
    else:
        from pyarrow.cffi import export_stream # type: ignore
        capsule = export_stream(reader)
    callback(capsule)

def arrow_batch(values, names):
    return pa.RecordBatchReader.from_batches(
        pa.schema(list(zip(names, [v.type for v in values]))),
        [pa.record_batch(values, names=names)],
    )

def _handle_query(sql, callback, **kwargs):
    cur = duckdb_con.cursor()
    query = sql.strip().lower().split(';')[0]

    if query.startswith("begin"):
        return callback("BEGIN", is_tag=True)
    if query.startswith("commit"):
        return callback("COMMIT", is_tag=True)
    if query.startswith("rollback"):
        return callback("ROLLBACK", is_tag=True)
    if query.startswith("discard all"):
        return callback("DISCARD ALL", is_tag=True)

    if query.startswith("select t.oid") and "from pg_type" in query and "hstore" in query:
        empty = [pa.array([], pa.int32()), pa.array([], pa.int32())]
        batch = arrow_batch(empty, ["oid", "typarray"])
        return send_reader(batch, callback)

    if query == "select pg_catalog.version()":
        batch = arrow_batch(
            [pa.array(["PostgreSQL 14.13"])],
            ["version"],
        )
        return send_reader(batch, callback)

    if query == "show transaction isolation level":
        batch = arrow_batch(
            [pa.array(["read committed"])],
            ["transaction_isolation"],
        )
        return send_reader(batch, callback)

    if query == "show standard_conforming_strings":
        batch = arrow_batch(
            [pa.array(["on"])],
            ["standard_conforming_strings"],
        )
        return send_reader(batch, callback)
    
    if query == "select current_schema()":
        batch = arrow_batch(
            [pa.array(["public"])],
            ["current_schema"],
        )
        return send_reader(batch, callback)

    if query.startswith("begin"):
        return callback("BEGIN", is_tag=True)
    if query.startswith("commit"):
        return callback("COMMIT", is_tag=True)
    if query.startswith("rollback"):
        return callback("ROLLBACK", is_tag=True)
    if query.startswith("discard all"):
        return callback("DISCARD ALL", is_tag=True)

    try:
        reader = cur.execute(sql).fetch_record_batch()
        send_reader(reader, callback)
    except Exception:
        logging.exception("error on executing query")
        batch = arrow_batch(
            [pa.array(["ERROR"]), pa.array(["unknown query"])],
            ["error", "message"],
        )
        send_reader(batch, callback)

def handle_query(sql, callback, **kwargs):
    executor.submit(_handle_query, sql, callback, **kwargs)

def main():
    global duckdb_con, executor
    duckdb_con = duckdb.connect("duckdata.db")
    duckdb_con.execute(
        """
        CREATE TABLE IF NOT EXISTS test_concurrency (
            id INTEGER,
            created_at TIMESTAMP,
            key TEXT,
            value TEXT
        )
    """
    )
    duckdb_con.execute("truncate table test_concurrency;")
    duckdb_con.execute(
        """
        INSERT INTO test_concurrency (id, created_at, key, value)
        VALUES
            (1, ?, 'alpha', 'value1'),
            (2, ?, 'beta', 'value2'),
            (3, ?, 'gamma', 'value3')
    """,
        [datetime.now(), datetime.now(), datetime.now()],
    )

    executor = ThreadPoolExecutor(max_workers=4)

    server = riffq.Server("127.0.0.1:5433")
    server.on_query(handle_query)
    server.start()

if __name__ == "__main__":
    main()



import multiprocessing
import socket
import time
import psycopg
import unittest
from helpers import _ensure_riffq_built
import pyarrow as pa


def _run_server(port: int):
    import riffq
    from riffq.helpers import to_arrow

    def handle_query(sql, callback, **kwargs):
        callback(to_arrow([{"name": "val", "type": "int"}], [[1]]))

    server = riffq.Server(f"127.0.0.1:{port}")
    server.register_database("mydb")
    server.register_schema("mydb", "myschema")
    server.register_table(
        "mydb",
        "myschema",
        "users",
        [{"id": {"type": "int", "nullable": False}}],
    )
    server.on_query(handle_query)
    server.start(catalog_emulation=True)


class RegisterCatalogTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _ensure_riffq_built()
        cls.port = 55440
        cls.proc = multiprocessing.Process(target=_run_server, args=(cls.port,), daemon=True)
        cls.proc.start()
        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()

    def test_registered_objects(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT datname FROM pg_catalog.pg_database WHERE datname='mydb'")
            self.assertEqual(cur.fetchone()[0], "mydb")
            cur.execute("SELECT nspname FROM pg_catalog.pg_namespace WHERE nspname='myschema'")
            self.assertEqual(cur.fetchone()[0], "myschema")
            cur.execute("SELECT relname FROM pg_catalog.pg_class WHERE relname='users'")
            self.assertEqual(cur.fetchone()[0], "users")
        conn.close()


if __name__ == "__main__":
    unittest.main()



import psycopg

with psycopg.connect("postgresql://user:password@localhost:5433/dbname") as conn:
    with conn.cursor() as cur:
        cur.execute("SELECT %s AS foo", (42,))
        print(cur.fetchone())
        cur.execute("SELECT %s AS foo", ('42',))
        print(cur.fetchone())


# from sqlalchemy import create_engine, text
#
#
# engine = create_engine("postgresql://myuser:mypassword@127.0.0.1:5433/mydb")
# conn = engine.connect()
# with engine.connect() as conn:
#     result = conn.execute(text("SELECT * FROM test;"))
#     rows = result.mappings().all()
#
# for row in rows:
#     print(dict(row))
#

# # Replace this with your actual server/port
# engine = create_engine("postgresql+psycopg2://user@localhost:5433/dbname")
# with engine.connect() as conn:
#     result = conn.execute(text("SELECT 1 + 2 AS bar"))
#     for row in result:
#         print(dict(row))
#
# with engine.connect() as conn:
#     stmt = text("SELECT :value AS foo")
#     result = conn.execute(stmt, {"value": 42})
#     for row in result:
#         print("row", row)
        # print(dict(row))
#



import multiprocessing
import socket
import subprocess
import sys
import time
from pathlib import Path

import psycopg
import unittest
import pyarrow as pa
from helpers import _ensure_riffq_built

def _run_server(port: int):
    import riffq
    import pyarrow as pa

    def send_batch(batch: pa.RecordBatch, callback):
        rdr = pa.RecordBatchReader.from_batches(batch.schema, [batch])
        # modern pyarrow
        if hasattr(rdr, "__arrow_c_stream__"):
            capsule = rdr.__arrow_c_stream__()
        else:            
            from pyarrow.cffi import export_stream  # type: ignore
            capsule = export_stream(rdr)
        callback(capsule)

    def handle_query(sql, callback, **kwargs):
        args = kwargs.get("query_args")
        sql_clean = sql.strip().lower()
        if sql_clean.startswith("begin"):
            return callback("BEGIN", is_tag=True)
        if sql_clean.startswith("commit"):
            return callback("COMMIT", is_tag=True)
        if sql_clean.startswith("rollback"):
            return callback("ROLLBACK", is_tag=True)
        if sql_clean.startswith("discard all"):
            return callback("DISCARD ALL", is_tag=True)
        if sql_clean == "select multi":
            batch = pa.record_batch(
                [
                    pa.array([1], pa.int64()),
                    pa.array(["foo"], pa.string()),
                    pa.array([3.14], pa.float64()),
                    pa.array([None], pa.int64()),
                ],
                names=["a", "b", "c", "d"],
            )
            return send_batch(batch, callback)
                    
        if sql_clean == "select bool":
            batch = pa.record_batch([pa.array([True], pa.bool_())], names=["flag"])
            return send_batch(batch, callback)

        if args:
            value = int(args[0])
        else:
            if sql_clean.startswith("select ") and sql_clean[7:].isdigit():
                value = int(sql_clean[7:])
            else:
                value = 1

        batch = pa.record_batch([pa.array([value], pa.int64())], names=["val"])
        send_batch(batch, callback)

    server = riffq.Server(f"127.0.0.1:{port}")
    server.on_query(handle_query)
    server.start()

class ServerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _ensure_riffq_built()
        cls.port = 55433
        cls.proc = multiprocessing.Process(
            target=_run_server, args=(cls.port,), daemon=True
        )
        cls.proc.start()

        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()

    def test_simple_query(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT 1")
            self.assertEqual(cur.fetchone()[0], 1)
        conn.close()

    def test_extended_query(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT %s", (42,))
            self.assertEqual(cur.fetchone()[0], 42)
        conn.close()

    def test_multiple_columns(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT multi")
            row = cur.fetchone()
            self.assertEqual(row, (1, "foo", 3.14, None))
            self.assertIsInstance(row[0], int)
            self.assertIsInstance(row[1], str)
            self.assertIsInstance(row[2], float)
            self.assertIsNone(row[3])

            names = [desc.name for desc in cur.description]
            self.assertEqual(names, ["a", "b", "c", "d"])
        conn.close()

    def test_bool_column(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT bool")
            row = cur.fetchone()
            self.assertEqual(row, (True,))
            self.assertIsInstance(row[0], bool)
            names = [desc.name for desc in cur.description]
            self.assertEqual(names, ["flag"])
        conn.close()

    def test_transaction_tags(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("BEGIN")
            self.assertEqual(cur.statusmessage, "BEGIN")
            cur.execute("ROLLBACK")
            self.assertEqual(cur.statusmessage, "ROLLBACK")
            cur.execute("BEGIN")
            cur.execute("COMMIT")
            self.assertEqual(cur.statusmessage, "COMMIT")
        conn.close()

if __name__ == "__main__":
    unittest.main()



import duckdb
import pyarrow as pa
import riffq
from riffq.helpers import to_arrow
import logging

def map_type(data_type: str) -> str:
    dt = data_type.upper()
    if "INT" in dt:
        return "int"
    if any(x in dt for x in ["DOUBLE", "DECIMAL", "REAL", "FLOAT"]):
        return "float"
    if "BOOL" in dt:
        return "bool"
    if dt == "DATE":
        return "date"
    if "TIMESTAMP" in dt or "DATETIME" in dt:
        return "datetime"
    return "str"

class Connection(riffq.BaseConnection):
    def _handle_query(self, sql, callback, **kwargs):
        cur = duckdb_con.cursor()
        sql = sql.strip().lower().split(';')[0]

        if sql.startswith("set"):
            return callback("SET", is_tag=True)

        if sql.startswith("begin"):
            return callback("BEGIN", is_tag=True)
        
        if sql.startswith("commit"):
            return callback("COMMIT", is_tag=True)
        
        if sql.startswith("rollback"):
            return callback("ROLLBACK", is_tag=True)
        
        if sql.startswith("discard all"):
            return callback("DISCARD ALL", is_tag=True)

        if sql == "select pg_catalog.version()":
            batch = self.arrow_batch(
                [pa.array(["PostgreSQL 14.13"])],
                ["version"],
            )
            return self.send_reader(batch, callback)

        if sql == "show transaction isolation level":
            batch = self.arrow_batch(
                [pa.array(["read committed"])],
                ["transaction_isolation"],
            )
            return self.send_reader(batch, callback)

        if sql == "show standard_conforming_strings":
            batch = self.arrow_batch(
                [pa.array(["read committed"])],
                ["transaction_isolation"],
            )
            return self.send_reader(batch, callback)
        
        if sql == "select current_schema()":
            batch = self.arrow_batch(
                [pa.array(["public"])],
                ["current_schema"],
            )
            return self.send_reader(batch, callback)

        try:
            reader = cur.execute(sql).fetch_record_batch()
            self.send_reader(reader, callback)
        except Exception as exc:
            logging.exception("error on executing query")
            batch = self.arrow_batch(
                [pa.array(["ERROR"]), pa.array([str(exc)])],
                ["error", "message"],
            )
            self.send_reader(batch, callback)

    def handle_query(self, sql, callback=callable, **kwargs):
        self.executor.submit(self._handle_query, sql, callback, **kwargs)

    def handle_auth(self, user, password, host, database=None, callback=callable):
        # return callback(user == "user" and password == "secret")
        callback(True)

def run_server(port: int):
    global duckdb_con
    duckdb_con = duckdb.connect()

    duckdb_con.execute("CREATE TABLE users(id INTEGER, name VARCHAR)")
    duckdb_con.execute("CREATE TABLE projects(id INTEGER, name VARCHAR)")
    duckdb_con.execute(
        "CREATE TABLE tasks(id INTEGER, project_id INTEGER, description VARCHAR)"
    )

    server = riffq.RiffqServer(f"127.0.0.1:{port}", connection_cls=Connection)
    server.set_tls("certs/server.crt", "certs/server.key")

    server._server.register_database("duckdb")

    tbls = duckdb_con.execute(
        "SELECT table_schema, table_name FROM information_schema.tables "
        "WHERE table_schema NOT IN ('pg_catalog','information_schema')"
    ).fetchall()

    for schema_name, table_name in tbls:
        server._server.register_schema("duckdb", schema_name)
        cols_info = duckdb_con.execute(
            "SELECT column_name, data_type, is_nullable FROM information_schema.columns "
            "WHERE table_schema=? AND table_name=?",
            (schema_name, table_name),
        ).fetchall()
        columns = []
        for col_name, data_type, is_nullable in cols_info:
            columns.append(
                {
                    col_name: {
                        "type": map_type(data_type),
                        "nullable": is_nullable.upper() == "YES",
                    }
                }
            )
        server._server.register_table("duckdb", schema_name, table_name, columns)

    server.start(catalog_emulation=True)

if __name__ == "__main__":
    run_server(port=5433)



import multiprocessing
import socket
import time
import unittest
import psycopg
from helpers import _ensure_riffq_built

import pyarrow as pa

def _run_server_catalog(port: int, enabled: bool):
    import riffq
    import pyarrow as pa

    def send_batch(batch: pa.RecordBatch, callback):
        rdr = pa.RecordBatchReader.from_batches(batch.schema, [batch])
        if hasattr(rdr, "__arrow_c_stream__"):
            capsule = rdr.__arrow_c_stream__()
        else:
            from pyarrow.cffi import export_stream  # type: ignore
            capsule = export_stream(rdr)
        callback(capsule)

    def handle_query(sql, callback, **kwargs):
        args = kwargs.get("query_args")
        sql_clean = sql.strip().lower()
        if sql_clean == "select multi":
            batch = pa.record_batch([
                pa.array([1], pa.int64()),
                pa.array(["foo"], pa.string()),
                pa.array([3.14], pa.float64()),
                pa.array([None], pa.int64()),
            ], names=["a", "b", "c", "d"])
            return send_batch(batch, callback)

        if sql_clean == "select bool":
            batch = pa.record_batch([pa.array([True], pa.bool_())], names=["flag"])
            return send_batch(batch, callback)

        if args:
            value = int(args[0])
        else:
            if sql_clean.startswith("select ") and sql_clean[7:].isdigit():
                value = int(sql_clean[7:])
            else:
                value = 1

        batch = pa.record_batch([pa.array([value], pa.int64())], names=["val"])
        send_batch(batch, callback)

    server = riffq.Server(f"127.0.0.1:{port}")
    server.on_query(handle_query)
    server.start(catalog_emulation=enabled)

class PgCatalogEnabledTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _ensure_riffq_built()
        cls.port = 55434
        cls.proc = multiprocessing.Process(target=_run_server_catalog, args=(cls.port, True), daemon=True)
        cls.proc.start()
        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()

    def test_catalog_query(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT datname FROM pg_catalog.pg_database LIMIT 1")
            row = cur.fetchone()
            self.assertIsInstance(row[0], str)
        conn.close()

    def test_catalog_pg_class_query(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT relname FROM pg_catalog.pg_class LIMIT 1")
            row = cur.fetchone()
            self.assertIsInstance(row[0], str)
        conn.close()

    def test_non_catalog_query(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT 42")
            self.assertEqual(cur.fetchone()[0], 42)
        conn.close()


class PgCatalogDisabledTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _ensure_riffq_built()
        cls.port = 55439
        cls.proc = multiprocessing.Process(target=_run_server_catalog, args=(cls.port, False), daemon=True)
        cls.proc.start()
        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()

    def test_pg_class_disabled(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT relname FROM pg_catalog.pg_class LIMIT 1")
            row = cur.fetchone()
            self.assertEqual(row[0], 1)
        conn.close()

if __name__ == "__main__":
    unittest.main()



import multiprocessing
import socket
import time
import psycopg
import unittest


def _run_server(port: int):
    import sys
    from pathlib import Path
    sys.path.insert(0, str(Path(__file__).resolve().parents[1]))
    from duckdb_pgcatalog.server import run_server
    run_server(port)


class DuckDbCatalogTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.port = 55441
        cls.proc = multiprocessing.Process(target=_run_server, args=(cls.port,), daemon=True)
        cls.proc.start()
        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()

    def test_catalog_entries(self):
        conn = psycopg.connect(f"postgresql://user:123@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT datname FROM pg_catalog.pg_database WHERE datname='duckdb'")
            self.assertEqual(cur.fetchone()[0], "duckdb")
            cur.execute("SELECT nspname FROM pg_catalog.pg_namespace WHERE nspname='main'")
            self.assertEqual(cur.fetchone()[0], "main")
            for table in ("users", "projects", "tasks"):
                cur.execute("SELECT relname FROM pg_catalog.pg_class WHERE relname=%s", (table,))
                self.assertEqual(cur.fetchone()[0], table)
        conn.close()


if __name__ == "__main__":
    unittest.main()



from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
import pyarrow as pa
from ._riffq import Server
from abc import ABCMeta, abstractmethod
import logging

logger = logging.getLogger('riffq:connection')

class BaseConnection(metaclass=ABCMeta):
    conn_id = None

    def __init__(self, conn_id, executor:ThreadPoolExecutor):
        self.conn_id = conn_id
        self.executor = executor

    def send_reader(self, reader, callback):
        if hasattr(reader, "__arrow_c_stream__"):
            capsule = reader.__arrow_c_stream__()
        else:
            # old pyarrow support
            from pyarrow.cffi import export_stream
            capsule = export_stream(reader)
        callback(capsule)

    def arrow_batch(self, values:list, names:list):
        return pa.RecordBatchReader.from_batches(
            pa.schema(list(zip(names, [v.type for v in values]))),
            [pa.record_batch(values, names=names)],
        )

    @abstractmethod
    def handle_auth(self, user, password, host, database=None, callback=callable):
        return callback(user == "user" and password == "secret")

    def handle_connect(self, ip, port, callback=callable):
        return callback(True)
    
    @abstractmethod
    def handle_query(self, sql, callback=callable, **kwargs):
        pass

    def handle_disconnect(self, ip, port, callback=callable):
        return callback(True)


class RiffqServer:
    def __init__(self, listen_addr, connection_cls=BaseConnection):
        # we aren't making the server class in rust/pyo3 subclassable
        # this way it's simpler
        self._server = Server(listen_addr)
        # TODO: max cpu
        self.executor = ThreadPoolExecutor(max_workers=4)
        self._server.on_authentication(self.handle_auth)
        self._server.on_query(self.handle_query)
        self._server.on_connect(self.handle_connect)
        self._server.on_disconnect(self.handle_disconnect)
        self.connections = {}
        self.connection_cls = connection_cls
    
    def set_tls(self, crt, key):
        self._server.set_tls(crt, key)

    def get_connection(self, conn_id) -> BaseConnection:
        conn = self.connections.get(conn_id, None)
        if not conn:
            conn = self.connection_cls(conn_id, self.executor)
            self.connections[conn_id] = conn
        return conn

    def handle_auth(self, conn_id, user, password, host, database=None, callback=callable):
        conn = self.get_connection(conn_id=conn_id)
        conn.handle_auth(user, password, host, database=database, callback=callback)

    def handle_connect(self, conn_id, ip, port, callback=callable):
        conn = self.get_connection(conn_id=conn_id)
        conn.handle_connect(ip, port, callback=callback)

    def handle_query(self, sql, callback, conn_id=None, **kwargs):
        conn = self.get_connection(conn_id=conn_id)
        conn.handle_query(sql, callback=callback, **kwargs)

    def handle_disconnect(self, conn_id, ip, port, callback=callable):
        conn = self.get_connection(conn_id=conn_id)
        conn.handle_disconnect(ip, port, callback=callback)
        try:
            del self.connections[conn_id]
        except KeyError:
            logger.exception("Connection disconnected but not in self.connections")

    def start(self, **kw):
        return self._server.start(**kw)



import multiprocessing
import socket
import subprocess
import sys
import time
import threading
from pathlib import Path

import psycopg
import unittest
from helpers import _ensure_riffq_built

import pyarrow as pa

def _run_server(port: int):
    import riffq
    from riffq.helpers import to_arrow

    def handle_query(sql, callback, **kw):
        if sql.strip().lower() == "select conn_id":
            capsule = to_arrow(
                [{"name": "id",  "type": "int"}],
                [[int(kw["connection_id"])]],
            )
        else:
            capsule = to_arrow(
                [{"name": "val", "type": "int"}],
                [[1]],
            )
        callback(capsule)

    server = riffq.Server(f"127.0.0.1:{port}")
    server.on_query(handle_query)
    server.start()


class ConnectionIdTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _ensure_riffq_built()
        cls.port = 55435
        cls.proc = multiprocessing.Process(target=_run_server, args=(cls.port,), daemon=True)
        cls.proc.start()
        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()

    def _fetch_id(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        with conn.cursor() as cur:
            cur.execute("SELECT conn_id")
            result = cur.fetchone()[0]
        conn.close()
        return result

    def test_connection_ids_increment(self):
        first = self._fetch_id()
        second = self._fetch_id()
        self.assertNotEqual(first, second)

    def test_parallel_connections_unique(self):
        results = []
        def worker():
            results.append(self._fetch_id())
        t1 = threading.Thread(target=worker)
        t2 = threading.Thread(target=worker)
        t1.start(); t2.start()
        t1.join(); t2.join()
        self.assertEqual(len(results), 2)
        self.assertEqual(len(set(results)), 2)


if __name__ == "__main__":
    unittest.main()



import logging
import threading

from riffq import riffq

def _handle_query(sql, callback, **kwargs):
    print("< received (python):", sql, kwargs)

    print("sql", sql.strip().lower())
    if sql.strip().lower() == "select pg_catalog.version()":
        result = (
            [
                {"name": "version", "type": "string"},
            ],
            [
                ["PostgreSQL 14.13 (Homebrew) on aarch64-apple-darwin23.4.0, compiled by Apple clang version 15.0.0 (clang-1500.3.9.4), 64-bit",]
            ]
        )

        callback(result)
        return

    result = (
        [ {"name": "error", "type": "str"}, {"name": "message", "type": "str"} ],
        [ ["ERROR", "unknown query"] ]
    )
    # result = (
    #     [ {"name": "col1", "type": "str"} ],
    #     [ ["hello"] ]
    # )

    callback(result)

def handle_query(sql, callback, **kwargs):
    try:
        _handle_query(sql, callback, **kwargs)
    except:
        logging.exception("exception on executing query")


if __name__ == "__main__":
    server = riffq.Server("127.0.0.1:5433")
    server.on_query(handle_query)
    server.start()



import unittest
import time
import signal
import threading
import duckdb
import logging
from datetime import datetime
from sqlalchemy import create_engine, text
from multiprocessing import Process
import riffq
import logging

from utils import wait_for_server

logging.basicConfig(level=logging.DEBUG)

def start_duckdb_server():
    from server_duckdb_arrow import main
    main()

def run_heavy_query():
    logging.info("sending long running query")
    engine = create_engine("postgresql://myuser:mypassword@127.0.0.1:5433/mydb")
    with engine.connect() as conn:
        conn.execute(text("""
            SELECT SUM(a.id * b.id)
            FROM range(0, 100000) AS a(id),
                 range(0, 10000) AS b(id)
        """))
    logging.info("finished long running query")


class TestDuckDBConcurrency(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server_proc = Process(target=start_duckdb_server)
        cls.server_proc.start()
        time.sleep(1.5)  # wait for server

    @classmethod
    def tearDownClass(cls):
        cls.server_proc.terminate()
        cls.server_proc.join()

    def test_concurrent_queries(self):
        engine = wait_for_server()

        fast_query_times = []

        heavy_proc = Process(target=run_heavy_query)
        heavy_proc.start()

        while heavy_proc.is_alive():
            try:
                with engine.connect() as conn:
                    logging.info("executing short running query")
                    result = conn.execute(text("SELECT * FROM test_concurrency;"))
                    rows = result.mappings().all()
                    logging.info("executed short running query")
                    if rows:
                        fast_query_times.append(time.time())
            except Exception:
                raise
            time.sleep(0.3)

        heavy_proc.join()
        print("fast query times", fast_query_times)

        self.assertGreaterEqual(len(fast_query_times), 2, "Expected at least 2 successful fast queries")


if __name__ == "__main__":
    unittest.main()



import logging
import threading

import riffq
import duckdb
import pandas
from concurrent.futures import ThreadPoolExecutor

def get_schema_from_duckdb(columns, types):
    duckdb_to_simple = {
        "BOOLEAN": "bool",
        "TINYINT": "int",
        "SMALLINT": "int",
        "INTEGER": "int",
        "BIGINT": "int",
        "HUGEINT": "int",
        "UTINYINT": "int",
        "USMALLINT": "int",
        "UINTEGER": "int",
        "UBIGINT": "int",
        "FLOAT": "float",
        "DOUBLE": "float",
        "DECIMAL": "float",
        "REAL": "float",
        "VARCHAR": "str",
        "BLOB": "str",
        "UUID": "str",
        "DATE": "date",
        "TIMESTAMP": "datetime",
        "TIMESTAMP_S": "datetime",
        "TIMESTAMP_MS": "datetime",
        "TIMESTAMP_NS": "datetime",
        "TIME": "str",  # optional: could be treated as 'time'
        "INTERVAL": "str",
    }
    schema = []
    for col, dtype in zip(columns, types):
        duck_type = str(dtype).upper()
        mapped = duckdb_to_simple.get(duck_type, "str")
        schema.append({"name": col, "type": mapped})
    return schema


def _handle_query(sql, callback, **kwargs):
    local_con = duckdb_con.cursor()
    print("< received (python):", sql)
    query = sql.strip().lower()

    if query.startswith("begin"):
        return callback("BEGIN", is_tag=True)
    if query.startswith("commit"):
        return callback("COMMIT", is_tag=True)
    if query.startswith("rollback"):
        return callback("ROLLBACK", is_tag=True)
    if query.startswith("discard all"):
        return callback("DISCARD ALL", is_tag=True)

    if query.startswith("select t.oid") and "from pg_type" in query and "hstore" in query:
        return callback(([{"name": "oid", "type": "int"}, {"name": "typarray", "type": "int"}], []))

    if query == "select pg_catalog.version()":
        result = (
            [
                {"name": "version", "type": "string"},
            ],
            [
                ["PostgreSQL 14.13 (Homebrew) on aarch64-apple-darwin23.4.0, compiled by Apple clang version 15.0.0 (clang-1500.3.9.4), 64-bit",]
            ]
        )

        return callback(result)


    if query == "show transaction isolation level":
        return callback(([
            {"name": "transaction_isolation", "type": "string"},
        ], [["read committed"]]))


    if query == "show standard_conforming_strings":
        return callback(([
            {"name": "standard_conforming_strings", "type": "string"},
        ], [["on"]]))

    if query == "select current_schema()":
        return callback(([
            {"name": "current_schema", "type": "string"},
        ], [["public"]]))


    if query.startswith("begin"):
        return callback("BEGIN", is_tag=True)
    if query.startswith("commit"):
        return callback("COMMIT", is_tag=True)
    if query.startswith("rollback"):
        return callback("ROLLBACK", is_tag=True)
    if query.startswith("discard all"):
        return callback("DISCARD ALL", is_tag=True)


    try:
        res = local_con.sql(sql)
        if res is None:
            callback(([], []))
            return
        schema = get_schema_from_duckdb(res.columns, res.types)
        callback((schema, res.fetchall()))
    except Exception as e:
        logging.exception("error on executing query")
        result = (
            [ {"name": "error", "type": "str"}, {"name": "message", "type": "str"} ],
            [ ["ERROR", "unknown query"] ]
        )
        callback(result)

def handle_query(sql, callback, **kwargs):
    def task():
        try:
            _handle_query(sql, callback)
        except Exception:
            logging.exception("exception on executing query")
    executor.submit(task)

def main():
    global duckdb_con, executor
    from datetime import datetime
    duckdb_con = duckdb.connect("duckdata.db", read_only=False)

    duckdb_con.execute("""
    CREATE TABLE IF NOT EXISTS test_concurrency (
        id INTEGER,
        created_at TIMESTAMP,
        key TEXT,
        value TEXT
    )
    """)
    duckdb_con.execute("truncate table test_concurrency;")
    duckdb_con.execute("""
    INSERT INTO test_concurrency (id, created_at, key, value)
    VALUES 
        (1, ?, 'alpha', 'value1'),
        (2, ?, 'beta', 'value2'),
        (3, ?, 'gamma', 'value3')
    """, [datetime.now(), datetime.now(), datetime.now()])

    executor = ThreadPoolExecutor(max_workers=4)

    server = riffq.Server("127.0.0.1:5433")
    server.on_query(handle_query)
    server.start()


if __name__ == "__main__":
    main()


import multiprocessing
import socket
import subprocess
import sys
import time
from pathlib import Path

import psycopg
import unittest
from helpers import _ensure_riffq_built

def _run_server(port: int):
    import riffq

    def handle_query(sql, callback, **kwargs):
        callback(([{"name": "val", "type": "int"}], [[1]]))

    def handle_connect(conn_id, ip, port, *, callback):
        callback(False)

    server = riffq.Server(f"127.0.0.1:{port}")
    server.on_query(handle_query)
    server.on_connect(handle_connect)
    server.start()


class OnConnectTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _ensure_riffq_built()
        cls.port = 55436
        cls.proc = multiprocessing.Process(target=_run_server, args=(cls.port,), daemon=True)
        cls.proc.start()
        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()

    def test_reject(self):
        with self.assertRaises(psycopg.OperationalError):
            psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")


if __name__ == "__main__":
    unittest.main()



import socket
import subprocess
import sys
import time
from pathlib import Path
import multiprocessing

import psycopg
import unittest
from helpers import _ensure_riffq_built


def _run_server(port: int, q):
    import riffq

    def handle_query(sql, callback, **kwargs):
        callback(([{"name": "val", "type": "int"}], [[1]]))

    def handle_disconnect(conn_id, ip, port):
        q.put(conn_id)

    server = riffq.Server(f"127.0.0.1:{port}")
    server.on_query(handle_query)
    server.on_disconnect(handle_disconnect)
    server.start()


class OnDisconnectTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _ensure_riffq_built()
        cls.port = 55437
        cls.queue = multiprocessing.Queue()
        cls.proc = multiprocessing.Process(target=_run_server, args=(cls.port, cls.queue), daemon=True)
        cls.proc.start()
        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()

    def test_disconnect_called(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db")
        conn.close()
        start = time.time()
        conn_id = None
        while time.time() - start < 5:
            if not self.queue.empty():
                conn_id = self.queue.get_nowait()
                break
            time.sleep(0.1)
        self.assertIsNotNone(conn_id, "disconnect callback not called")
        self.assertIsInstance(conn_id, int)


if __name__ == "__main__":
    unittest.main()



from sqlalchemy import create_engine, text
import time

def wait_for_server(port: int = 5433):
    engine = create_engine(f"postgresql://myuser:mypassword@127.0.0.1:{port}/mydb")
    import sqlalchemy.exc
    cnt = 0
    while True:
        try:
            with engine.connect() as conn:
                result = conn.execute(text("SELECT * FROM test_concurrency;"))
                print("connected")
                break
        except sqlalchemy.exc.OperationalError:
            print("waiting for server to be online")
            time.sleep(1)
            cnt += 1
        if cnt > 10:
            raise Exception("couldnt spawn the server")
    
    return engine




import pyarrow as pa

_type = {
    "int": pa.int64(),
    "float": pa.float64(),
    "bool": pa.bool_(),
    "str": pa.utf8(),
    "string": pa.utf8(),
    "date": pa.date32(),
    "datetime": pa.timestamp("us"),
}

def to_arrow(schema_desc, rows):
    arrays = []
    for col_ix, col in enumerate(schema_desc):
        ty  = _type[col["type"]]
        arr = pa.array([r[col_ix] for r in rows], type=ty)
        arrays.append(arr)
    batch  = pa.RecordBatch.from_arrays(arrays, names=[c["name"] for c in schema_desc])
    reader = pa.RecordBatchReader.from_batches(batch.schema, [batch])
    return reader.__arrow_c_stream__()         



import multiprocessing
import socket
import subprocess
import sys
import time
from pathlib import Path
import tempfile

import psycopg
import unittest
from helpers import _ensure_riffq_built

def _run_server_tls(port: int, cert: str, key: str):
    import riffq
    from riffq.helpers import to_arrow

    def handle_query(sql, callback, **kwargs):
        args = kwargs.get("query_args")

        sql_clean = sql.strip().lower()
        if sql_clean == "select 1":
            callback(to_arrow([{"name": "val", "type": "int"}], [[1]]) )
            return
        if args:
            value = int(args[0])
        elif sql_clean.startswith("select ") and sql_clean[7:].isdigit():
            value = int(sql_clean[7:])
        else:
            value = 1
        callback(to_arrow([{"name": "val", "type": "int"}], [[value]]))

    server = riffq.Server(f"127.0.0.1:{port}")
    server.on_query(handle_query)
    server.set_tls(cert, key)
    server.start(tls=True)


class ServerTLSTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _ensure_riffq_built()
        cls.port = 55434
        cls.tmp = tempfile.TemporaryDirectory()
        cert = Path(cls.tmp.name) / "server.crt"
        key = Path(cls.tmp.name) / "server.key"
        subprocess.check_call([
            "openssl", "req", "-newkey", "rsa:2048", "-nodes",
            "-keyout", str(key), "-x509", "-days", "1",
            "-out", str(cert), "-subj", "/CN=localhost"
        ])
        cls.proc = multiprocessing.Process(target=_run_server_tls, args=(cls.port, str(cert), str(key)), daemon=True)
        cls.proc.start()

        start = time.time()
        while time.time() - start < 10:
            with socket.socket() as sock:
                if sock.connect_ex(("127.0.0.1", cls.port)) == 0:
                    break
            time.sleep(0.1)
        else:
            cls.proc.terminate()
            cls.proc.join()
            raise RuntimeError("Server did not start")

    @classmethod
    def tearDownClass(cls):
        cls.proc.terminate()
        cls.proc.join()
        cls.tmp.cleanup()

    def test_tls_query(self):
        conn = psycopg.connect(f"postgresql://user@127.0.0.1:{self.port}/db", sslmode="require")
        with conn.cursor() as cur:
            cur.execute("SELECT 1")
            self.assertEqual(cur.fetchone()[0], 1)
        conn.close()


if __name__ == "__main__":
    unittest.main()



import logging
import threading
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor
import pandas
import polars as pl
import riffq

executor = ThreadPoolExecutor(max_workers=4)

def get_schema_from_polars(df: pl.DataFrame):
    polars_to_simple = {
        pl.Boolean: "bool",
        pl.Int8: "int",
        pl.Int16: "int",
        pl.Int32: "int",
        pl.Int64: "int",
        pl.UInt8: "int",
        pl.UInt16: "int",
        pl.UInt32: "int",
        pl.UInt64: "int",
        pl.Float32: "float",
        pl.Float64: "float",
        pl.String: "str",
        pl.Binary: "str",
        pl.Date: "date",
        pl.Datetime: "datetime",
        pl.Time: "str",
        pl.Duration: "str",
    }

    schema = []
    for col in df.schema:
        dtype = df.schema[col]
        mapped = polars_to_simple.get(dtype, "str")
        schema.append({"name": col, "type": mapped})
    return schema

def run_heavy_polars_query():
    df = heavy_df
    # This forces heavy CPU math in native Polars (Rust)
    return df.lazy().with_columns([
        (pl.col("x") ** 0.5).log().sin().cos().tan().exp().alias("result")
    ]).select([
        pl.col("result").sum()
    ]).collect()

def _handle_query(sql, callback, **kwargs):
    print("< received (python):", sql)
    query = sql.strip().lower()

    if query.startswith("begin"):
        return callback("BEGIN", is_tag=True)
    if query.startswith("commit"):
        return callback("COMMIT", is_tag=True)
    if query.startswith("rollback"):
        return callback("ROLLBACK", is_tag=True)
    if query.startswith("discard all"):
        return callback("DISCARD ALL", is_tag=True)

    if query.startswith("select t.oid") and "from pg_type" in query and "hstore" in query:
        return callback(([{"name": "oid", "type": "int"}, {"name": "typarray", "type": "int"}], []))

    if query == "select pg_catalog.version()":
        return callback(([{"name": "version", "type": "string"}],
                         [["PostgreSQL 14.13"]]))

    if query == "show transaction isolation level":
        return callback(([{"name": "transaction_isolation", "type": "string"}],
                         [["read committed"]]))

    if query == "show standard_conforming_strings":
        return callback(([{"name": "standard_conforming_strings", "type": "string"}],
                         [["on"]]))

    if query == "select current_schema()":
        return callback(([{"name": "current_schema", "type": "string"}],
                         [["public"]]))


    if query.startswith("begin"):
        return callback("BEGIN", is_tag=True)
    if query.startswith("commit"):
        return callback("COMMIT", is_tag=True)
    if query.startswith("rollback"):
        return callback("ROLLBACK", is_tag=True)
    if query.startswith("discard all"):
        return callback("DISCARD ALL", is_tag=True)


    if query.strip().lower().replace(';', '') == "select heavy_query":
        df = run_heavy_polars_query()
        schema = get_schema_from_polars(df)
        rows = df.to_numpy().tolist()
        return callback((schema, rows))

    try:
        sql_ctx = pl.SQLContext()
        sql_ctx.register("test_concurrency", test_concurrency_df)
        df = sql_ctx.execute(sql).collect()
        schema = get_schema_from_polars(df)
        rows = df.to_numpy().tolist()
        callback((schema, rows))
    except Exception:
        logging.exception("error on executing query")
        result = (
            [{"name": "error", "type": "str"}, {"name": "message", "type": "str"}],
            [["ERROR", "unknown query"]],
        )
        callback(result)

def handle_query(sql, callback, **kwargs):
    def task():
        try:
            _handle_query(sql, callback)
        except Exception:
            logging.exception("exception on executing query")
    executor.submit(task)

def main():
    global test_concurrency_df, heavy_df

    heavy_df = pl.DataFrame({
        # reduced range to keep memory usage reasonable while
        # still providing a noticeably heavy query
        "x": list(range(10_000_000))
    })

    test_concurrency_df = pl.DataFrame({
        "id": [1, 2, 3],
        "created_at": [datetime.now()] * 3,
        "key": ["alpha", "beta", "gamma"],
        "value": ["value1", "value2", "value3"]
    })
    server = riffq.Server("127.0.0.1:5434")
    server.on_query(handle_query)
    server.start()

if __name__ == "__main__":
    main()



from ._riffq import Server  # Rust class
from . import connection 
from .connection import BaseConnection, RiffqServer

BaseConnection = connection.BaseConnection
RiffqServer    = connection.RiffqServer